import { IntermediateBindingStatus } from "../logic/character";
import { BindingType, BodyPartType } from "../logic/definitions";
import { changeImageColor, createImage } from "./images";

export type ImageColors = { [colorName: string]: string };

const definitions: ImageLoopCriteriaDefinition[] = [
  {
    criteria: (parts) =>
      parts.some((p) => p.bodyPart === "eyes" && p.isTied) &&
      parts.some((p) => p.bodyPart === "mouth" && p.isTied),
    overridableLayers: {
      hair: createStdLayerDefinition("hair_all", "hair"),
    },
    stackableLayers: [],
  },
  hairCriteriaDefinition("eyes", "tape", "tape_eyes", "hair_eyes", "tape"),
  hairCriteriaDefinition("mouth", "cloth", "cloth_gag", "hair_mouth"),
  hairCriteriaDefinition("mouth", "ballgag", "ballgag", "hair_mouth"),
  hairCriteriaDefinition("mouth", "tape", "tape_mouth", "hair_mouth", "tape"),
  simpleCriteriaDefinition("arms", "straps", "straps_chest"),
  simpleCriteriaDefinition("legs", "straps", "straps_legs"),
  simpleCriteriaDefinition("ankles", "straps", "straps_ankles"),
  simpleCriteriaDefinition("arms", "tape", "tape_chest", "tape"),
  simpleCriteriaDefinition("legs", "tape", "tape_legs", "tape"),
  simpleCriteriaDefinition("ankles", "tape", "tape_ankles", "tape"),
  {
    criteria: (_) => true,
    overridableLayers: {
      hair: createStdLayerDefinition("hair_free", "hair"),
      top_lines: createStdLayerDefinition("top_lines", "top_lines"),
      top: createStdLayerDefinition("top", "top"),
      pants_lines: createStdLayerDefinition("pants_lines", "pants_lines"),
      pants: createStdLayerDefinition("pants", "pants"),
      skin_lines: createStdLayerDefinition("skin_lines", "skin_lines"),
      skin: createStdLayerDefinition("skin", "skin"),
    },
    stackableLayers: [],
  },
];

let imageLoops: Map<string, ImageLoopCriteria[]> = new Map();

function getCacheKeyFrom(colors: ImageColors) {
  return Object.keys(colors)
    .map((k) => `${k}-${colors[k]}`)
    .join(";");
}

export async function getCachedImages(
  colors: ImageColors,
  progress: (progress: number, maxProgress: number) => void
): Promise<ImageLoopCriteria[]> {
  const colorKeyCache = getCacheKeyFrom(colors);
  const cachedEntry = imageLoops.get(colorKeyCache);
  if (cachedEntry) return cachedEntry;
  const generator = async function* (): AsyncGenerator<ImageLoopCriteria> {
    let counter = 0;
    for (const entryDefinition of definitions) {
      const loadedEntry: ImageLoopCriteria = {
        criteria: entryDefinition.criteria,
        overridableLayers: Object.fromEntries(
          await Promise.all(
            Object.entries(entryDefinition.overridableLayers).map(
              async ([k, s]) => [
                k,
                {
                  frames: await Promise.all(
                    s.frameFiles.map(
                      async (f) => await loadFrame(f, colors, s.color)
                    )
                  ),
                },
              ]
            )
          )
        ),
        stackableLayers: await Promise.all(
          entryDefinition.stackableLayers.map(async (s) => ({
            frames: await Promise.all(
              s.frameFiles.map(async (f) => await loadFrame(f, colors, s.color))
            ),
          }))
        ),
      };
      counter++;
      progress(counter, definitions.length);
      yield loadedEntry;
    }
  };
  const result: ImageLoopCriteria[] = [];
  const generatorEntries = generator();
  for await (const entry of generatorEntries) {
    result.push(entry);
  }
  imageLoops.set(colorKeyCache, result);
  return result;
}

const cachedColoredImages: {
  [file: string]: {
    [color: string]: HTMLImageElement;
  };
} = {};
export async function getCachedReskinnedImage(
  image: ImageFrame,
  color: string
) {
  if (!cachedColoredImages[image.file]) cachedColoredImages[image.file] = {};
  if (!cachedColoredImages[image.file][color])
    cachedColoredImages[image.file][color] = await changeImageColor(
      image.img,
      color
    );
  return cachedColoredImages[image.file][color];
}

function hairCriteriaDefinition<T extends BodyPartType>(
  bodyPart: T,
  bindingType: BindingType<T>,
  image: string,
  hairImage: string,
  color?: string
): ImageLoopCriteriaDefinition {
  return {
    criteria: (parts) =>
      parts.some(
        (p) => p.bodyPart === bodyPart && p.binding === bindingType && p.isTied
      ),
    overridableLayers: {
      hair: createStdLayerDefinition(hairImage, "hair"),
    },
    stackableLayers: [createStdLayerDefinition(image, color)],
  };
}

function simpleCriteriaDefinition<T extends BodyPartType>(
  bodyPart: T,
  bindingType: BindingType<T>,
  image: string,
  color?: string
): ImageLoopCriteriaDefinition {
  return {
    criteria: (parts) =>
      parts.some(
        (p) => p.bodyPart === bodyPart && p.binding === bindingType && p.isTied
      ),
    overridableLayers: {},
    stackableLayers: [createStdLayerDefinition(image, color)],
  };
}

function createStdLayerDefinition(
  fileName: string,
  color?: string
): LayerContentDefinition {
  return {
    color,
    frameFiles: [
      `f1_${fileName}.png`,
      `f2_${fileName}.png`,
      `f3_${fileName}.png`,
      `f2_${fileName}.png`,
    ],
  };
}

async function loadFrame(
  name: string,
  colors: ImageColors,
  color?: string
): Promise<ImageFrame> {
  let f1 = await createImage(name);
  if (color && colors[color]) {
    f1 = await getCachedReskinnedImage({ file: name, img: f1 }, colors[color]);
  }
  return { file: name, img: f1 };
}

export const overridableLayers = [
  "hair",
  "skin",
  "pants",
  "top",
  "skin_lines",
  "pants_lines",
  "top_lines",
] as const;

export type OverridableLayerName = typeof overridableLayers[number];

export type ImageFrame = { file: string; img: HTMLImageElement };
export type LayerContent = { frames: ImageFrame[] };

export type ImageLoopCriteria = {
  criteria: (parts: IntermediateBindingStatus<BodyPartType>[]) => boolean;
  overridableLayers: { [name in OverridableLayerName]?: LayerContent };
  stackableLayers: LayerContent[];
};

type LayerContentDefinition = {
  color?: string;
  frameFiles: string[];
};

type ImageLoopCriteriaDefinition = {
  criteria: (parts: IntermediateBindingStatus<BodyPartType>[]) => boolean;
  overridableLayers: {
    [name in OverridableLayerName]?: LayerContentDefinition;
  };
  stackableLayers: LayerContentDefinition[];
};
