import { BodyPartBindingStatus, CharacterBodyPartStatus } from "./character";
import {
  BodyPartType,
  bodyPartBindingDefinitions,
  bodyPartList,
  BindingType,
  getModifiers,
} from "./definitions";
import { ActiveEffects } from "./effects";
import { assertUnreachable, getRandom, getRandomArrayItem } from "./utils";

export function getNewTiedBodyParts(
  currentStatus: CharacterBodyPartStatus,
  difficulty: number,
  activeEffects: ActiveEffects
): CharacterBodyPartStatus {
  if (!bodyPartList.some((b) => !currentStatus[b])) {
    // The character is fully tied. Enhance a single binding by the difficulty level.
    const tightenLevel = Math.ceil(
      (1.5 * getRandom() * difficulty) /
        activeEffects.decreaseNewBindingsDifficulty
    );
    const targetBodyPart = getRandomArrayItem(bodyPartList);
    const targetStatus = currentStatus[targetBodyPart];
    if (targetStatus) {
      targetStatus.level += tightenLevel;
    }
    return currentStatus;
  }

  const wristsTightness = Math.ceil(
    (1.5 * getRandom() * difficulty) /
      activeEffects.decreaseNewBindingsDifficulty
  );
  let addedBindings = 0;
  const result: CharacterBodyPartStatus = currentStatus;
  if (!result.wrists) {
    result.wrists = {
      binding: getRandomBinding("wrists"),
      level: wristsTightness,
      struggledPercentage: 0,
    };
    addedBindings++;
  }

  let remainingBodyParts = bodyPartList
    .filter((b) => !result[b])
    .filter((b) => b !== "wrists");
  while (remainingBodyParts.length > 0) {
    if (
      getFullEquivalentLevel(result, activeEffects) > difficulty &&
      addedBindings > 0
    ) {
      break;
    }
    addedBindings++;
    const bodyPart = getRandomArrayItem(remainingBodyParts);
    remainingBodyParts = remainingBodyParts.filter((b) => b !== bodyPart);
    const tightness = Math.ceil(
      (1.5 * getRandom() * difficulty) /
        activeEffects.decreaseNewBindingsDifficulty
    );
    if (bodyPart === "ankles") {
      result[bodyPart] = generateBodyPartStatus(bodyPart, tightness);
    } else if (bodyPart === "arms") {
      result[bodyPart] = generateBodyPartStatus(bodyPart, tightness);
    } else if (bodyPart === "eyes") {
      result[bodyPart] = generateBodyPartStatus(bodyPart, tightness);
    } else if (bodyPart === "hands") {
      result[bodyPart] = generateBodyPartStatus(bodyPart, tightness);
    } else if (bodyPart === "legs") {
      result[bodyPart] = generateBodyPartStatus(bodyPart, tightness);
    } else if (bodyPart === "mouth") {
      result[bodyPart] = generateBodyPartStatus(bodyPart, tightness);
    } else {
      assertUnreachable(bodyPart);
    }
  }
  return result;
}

export function generateBodyPartStatus<T extends BodyPartType>(
  bodyPart: T,
  level: number
): BodyPartBindingStatus<T> {
  return {
    binding: getRandomBinding(bodyPart),
    level,
    struggledPercentage: 0,
  };
}

export function getRandomBinding<T extends BodyPartType>(
  bodyPart: T
): BindingType<T> {
  const bindings = bodyPartBindingDefinitions[bodyPart].bindings;
  const keys = Object.keys(bindings) as BindingType<T>[];
  return getRandomArrayItem(keys);
}

export function removeBinding(
  bodyParts: CharacterBodyPartStatus,
  bodyPart: BodyPartType
):
  | {
      bodyPart: BodyPartType;
      binding: { binding: BindingType<BodyPartType>; level: number };
      after: CharacterBodyPartStatus;
    }
  | undefined {
  const binding = bodyParts[bodyPart];
  if (!binding) return;
  bodyParts[bodyPart] = undefined;
  return { bodyPart, binding, after: bodyParts };
}

export function removeRandomBinding(bodyParts: CharacterBodyPartStatus):
  | {
      bodyPart: BodyPartType;
      binding: { binding: BindingType<BodyPartType>; level: number };
      after: CharacterBodyPartStatus;
    }
  | undefined {
  const remainingBodyParts: BodyPartType[] = (
    Object.keys(bodyParts) as BodyPartType[]
  ).filter((b) => b !== "wrists") || ["wrists"];
  if (remainingBodyParts.length === 0) return;
  const selectedBodyPart = getRandomArrayItem(remainingBodyParts);
  return removeBinding(bodyParts, selectedBodyPart);
}

export function getBodyPartEquivalentLevel(
  bodyParts: CharacterBodyPartStatus,
  bodyPart: BodyPartType,
  activeEffects: ActiveEffects
): number {
  const content = bodyParts[bodyPart];
  if (!content) return 0;

  let difficultyRatio = 1.0;
  const definition = getModifiers(bodyPart);
  for (const bodyPart of bodyPartList) {
    const data = definition[bodyPart];
    if (!data) continue;
    if (!bodyParts[bodyPart]) continue;
    difficultyRatio *= data.ratio;
  }
  if (difficultyRatio <= 0.0) return Number.POSITIVE_INFINITY;
  const baseEquivalentLevel =
    content.level * (1.0 - content.struggledPercentage);
  return Math.ceil(
    baseEquivalentLevel /
      difficultyRatio /
      activeEffects.increaseStrugglingStrength
  );
}

export function getFullEquivalentLevel(
  bodyParts: CharacterBodyPartStatus,
  activeEffects: ActiveEffects
) {
  let level = 0;
  for (const bodyPart of bodyPartList) {
    const equivalentLevel = getBodyPartEquivalentLevel(
      bodyParts,
      bodyPart,
      activeEffects
    );
    level += equivalentLevel;
  }
  return level;
}
