import { getBodyPartEquivalentLevel } from "./bindings";
import { BindingType, bodyPartList, BodyPartType } from "./definitions";
import { getBodyPartDisplayName, getBindingDisplayName } from "./display";
import { ActiveEffects } from "./effects";
import { getRandom } from "./utils";

export type BodyPartBindingStatus<Part extends BodyPartType> = {
  binding: BindingType<Part>;
  level: number;
  struggledPercentage: number;
};

export type CharacterBodyPartStatus = {
  [Part in BodyPartType]?: BodyPartBindingStatus<Part>;
};

export type CharacterStatus = {
  bodyParts: CharacterBodyPartStatus;
};

export function isTied(character: CharacterStatus): boolean {
  return Object.values(character.bodyParts).some((t) => !!t);
}

export function getCurrentTiedBodyParts(
  character: CharacterStatus
): BodyPartType[] {
  const result: BodyPartType[] = [];
  for (const part of bodyPartList) {
    const content = character.bodyParts[part];
    if (content) result.push(part);
  }
  return result;
}

export type IntermediateBindingStatus<T extends BodyPartType> = {
  bodyPart: BodyPartType;
  binding?: BindingType<T>;
  isTied: boolean;
  struggledPercentage: number;
  bodyPartName: string;
  bindingDisplayName: string;
  baseLevel: number;
  level: number;
};

export function getCharacterStatusPerBodyPart(
  character: CharacterStatus,
  activeEffects: ActiveEffects
): IntermediateBindingStatus<BodyPartType>[] {
  const result: IntermediateBindingStatus<BodyPartType>[] = [];
  for (const part of bodyPartList) {
    const content = character.bodyParts[part];
    if (content) {
      result.push({
        bodyPart: part,
        binding: content.binding,
        isTied: true,
        struggledPercentage: content.struggledPercentage,
        bodyPartName: getBodyPartDisplayName(part, { capitalized: true }),
        bindingDisplayName: getBindingDisplayName(part, content.binding, {
          capitalized: true,
        }),
        baseLevel: content.level,
        level: getBodyPartEquivalentLevel(
          character.bodyParts,
          part,
          activeEffects
        ),
      });
    } else {
      result.push({
        bodyPart: part,
        isTied: false,
        struggledPercentage: 1.0,
        bodyPartName: getBodyPartDisplayName(part, { capitalized: true }),
        bindingDisplayName: "Free",
        baseLevel: 0,
        level: 0,
      });
    }
  }
  return result;
}

export function getMaxTimeoutFrom(character: CharacterStatus): number {
  let difficulty = 1;
  for (const part of bodyPartList) {
    const content = character.bodyParts[part];
    if (!content) continue;
    difficulty += content.level;
  }

  return difficulty * (1 + getRandom()) * 1000;
}
