import { ActiveEffects } from "./effects";
import { LootItem } from "./loot";
import {
  assertUnreachable,
  getRandomArrayItem,
  getRandomItemInWeightedArray,
  WeightedArray,
} from "./utils";

const CURRENT_ROOM_LEVEL = 5;
export type RoomPosition = { x: number; y: number };

export type RoomContent =
  | { type: "cleared" }
  | { type: "lootable"; difficulty: number }
  | { type: "looted" }
  | { type: "trap"; difficulty: number };

const roomExitDirectionList = ["n", "s", "e", "w"] as const;
type RoomExitDirection = typeof roomExitDirectionList[number];
export type RoomType = RoomPosition & {
  paths: { [key in RoomExitDirection]?: boolean };
  explorationLevel: number;
  contents?: RoomContent;
};

export type RoomList = Map<string, RoomType>;

export function getCoordinates(position: RoomPosition): string {
  return `${position.x},${position.y}`;
}

function chooseRandomDirectionsFrom(
  directions: readonly RoomExitDirection[],
  percent: number
): RoomExitDirection[] {
  const pickedDirections: Set<RoomExitDirection> = new Set();
  for (let i = 0; i < directions.length * percent; ++i) {
    const pickedDirection = getRandomArrayItem(directions);
    pickedDirections.add(pickedDirection);
  }
  return [...pickedDirections.values()];
}

export function pickRandomDirection(
  position: RoomPosition,
  rooms: RoomList
): "n" | "s" | "e" | "w" {
  const existingRoom = rooms.get(getCoordinates(position));
  if (!existingRoom) return "n";
  const coords = [
    { x: position.x, y: position.y - 1, direction: "n" },
    { x: position.x, y: position.y + 1, direction: "s" },
    { x: position.x - 1, y: position.y, direction: "w" },
    { x: position.x + 1, y: position.y, direction: "e" },
  ] as const;
  const possibleCoords = coords.filter((c) => existingRoom.paths[c.direction]);
  const unexploredCoords: typeof possibleCoords = [];
  const weightedCoords: WeightedArray<typeof possibleCoords[number]> = [];
  for (const coord of possibleCoords) {
    const room = rooms.get(getCoordinates(coord));
    if (!room) continue;
    if (!room.contents) {
      unexploredCoords.push(coord);
      continue;
    }
    const exploration = room.explorationLevel || 0;
    let weight = 1 / (2 * exploration + 1);
    if (room.contents.type === "trap") weight /= 10;
    weightedCoords.push({ weight, item: coord });
  }
  if (unexploredCoords.length > 0) {
    return getRandomArrayItem(unexploredCoords).direction;
  }
  const pickedCoord = getRandomItemInWeightedArray(weightedCoords);
  return pickedCoord.direction;
}

export function moveToNewPosition(
  position: RoomPosition,
  rooms: RoomList,
  direction: "n" | "s" | "e" | "w"
): RoomPosition {
  const existingRoom = rooms.get(getCoordinates(position));
  if (!existingRoom) {
    console.warn(
      `Could not move from position ${getCoordinates(
        position
      )}: No existing room at the given position`
    );
    return position;
  }
  const coords = [
    { x: position.x, y: position.y - 1, direction: "n" },
    { x: position.x, y: position.y + 1, direction: "s" },
    { x: position.x - 1, y: position.y, direction: "w" },
    { x: position.x + 1, y: position.y, direction: "e" },
  ] as const;
  const possibleCoords = coords.filter((c) => existingRoom.paths[c.direction]);
  const targetCoord =
    possibleCoords.find((c) => c.direction === direction) ||
    getRandomArrayItem(possibleCoords);
  return { x: targetCoord.x, y: targetCoord.y };
}

export function createStartingArea(): {
  position: RoomPosition;
  rooms: RoomList;
} {
  const position: RoomPosition = {
    x: 0,
    y: 0,
  };
  const starterRoom: RoomType = {
    x: position.x,
    y: position.y,
    paths: {},
    explorationLevel: CURRENT_ROOM_LEVEL,
    contents: { type: "cleared" },
  };

  for (const direction of chooseRandomDirectionsFrom(
    roomExitDirectionList,
    1.0
  )) {
    starterRoom.paths[direction] = true;
  }

  const roomList: RoomList = new Map();
  roomList.set(getCoordinates(position), starterRoom);
  return {
    position: { x: position.x, y: position.y },
    rooms: createNewRoomsAround({ x: position.x, y: position.y }, roomList),
  };
}

export function createNewRoomsAround(
  roomPosition: { x: number; y: number },
  rooms: RoomList
): RoomList {
  const existingRoom = rooms.get(getCoordinates(roomPosition));
  if (!existingRoom) {
    console.warn(
      `Could not create rooms around ${getCoordinates(
        roomPosition
      )}: No existing rooms at the given position`
    );
    return rooms;
  }
  const coords = [
    { x: roomPosition.x, y: roomPosition.y - 1, direction: "n" },
    { x: roomPosition.x, y: roomPosition.y + 1, direction: "s" },
    { x: roomPosition.x - 1, y: roomPosition.y, direction: "w" },
    { x: roomPosition.x + 1, y: roomPosition.y, direction: "e" },
  ] as const;
  for (const coord of coords) {
    if (!existingRoom.paths[coord.direction]) continue;
    if (rooms.has(getCoordinates(coord))) continue;
    const newRoom = createNewRoom({ x: coord.x, y: coord.y }, rooms);
    rooms.set(getCoordinates(coord), newRoom);
  }
  return rooms;
}

function createNewRoom(
  position: RoomPosition,
  existingRooms: RoomList
): RoomType {
  let newDirections = roomExitDirectionList.map((d) => d);
  const existingDirections: RoomExitDirection[] = [];
  function checkRoom(
    x: number,
    y: number,
    roomDirection: RoomExitDirection,
    newDirection: RoomExitDirection
  ) {
    const room = existingRooms.get(getCoordinates({ x, y }));
    if (!room) return;
    newDirections = newDirections.filter((d) => d !== newDirection);
    if (room.paths[roomDirection]) existingDirections.push(newDirection);
  }
  checkRoom(position.x, position.y - 1, "s", "n");
  checkRoom(position.x, position.y + 1, "n", "s");
  checkRoom(position.x - 1, position.y, "e", "w");
  checkRoom(position.x + 1, position.y, "w", "e");
  const room: RoomType = {
    x: position.x,
    y: position.y,
    paths: {},
    explorationLevel: 0,
  };
  for (const direction of existingDirections) {
    room.paths[direction] = true;
  }
  for (const direction of chooseRandomDirectionsFrom(newDirections, 0.5)) {
    room.paths[direction] = true;
  }
  return room;
}
export function decreaseExploreLevels(
  position: RoomPosition,
  rooms: RoomList
): RoomList {
  for (const room of rooms.values()) {
    if (position.x === room.x && position.y === room.y) {
      room.explorationLevel = CURRENT_ROOM_LEVEL;
      continue;
    }
    if (room.explorationLevel <= 0) continue;
    room.explorationLevel--;
  }
  return rooms;
}

export function generateRoomContents(
  level: number,
  effects: ActiveEffects
): RoomContent {
  const choices: WeightedArray<"cleared" | "lootable" | "trap"> = [
    { weight: 7, item: "cleared" },
    { weight: 1, item: "lootable" },
    { weight: 1, item: "trap" },
  ];
  const choice = getRandomItemInWeightedArray(choices);
  if (choice === "cleared") {
    return { type: "cleared" };
  }
  if (choice === "trap") {
    return { type: "trap", difficulty: level };
  }
  if (choice === "lootable") {
    return {
      type: "lootable",
      difficulty: level * effects.increaseExplorationLootLevel,
    };
  }
  assertUnreachable(choice);
}
